#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <unistd.h>
//#include <signal.h>
//#include <time.h>
//#include <sys/types.h>
//#include <sys/sysinfo.h>
//#include <sys/stat.h>
#include <stdint.h>
//#include <syslog.h>
#include <dirent.h>
#include <time.h>

//#define DEBUG_CODE
//#define DEBUG_HOST      "192.168.0.107"

// Maximum days in date range to store data
#define MAX_RANGE 		31
// Maximum # of clients to keep track of
#define MAX_CLIENTCOUNT 100

#define INCOMING_DATA	0
#define OUTGOING_DATA	1

static uint64_t start_date = 0;
static uint64_t end_date   = 0;

typedef struct
{
    char ip[16];
    double bytes_in[MAX_RANGE];
    double bytes_out[MAX_RANGE];
    double previous_bytes_in[MAX_RANGE];
    double previous_bytes_out[MAX_RANGE];

#ifdef DEBUG_CODE
    char previous_bytes_in_str[1024];
    char previous_bytes_out_str[1024];
#endif

} data_t;

typedef struct
{
    char sPackets[9];
    char sBytes[9];
    char sProt[5];
    char sOp[5];
    char sIn[5];
    char sOut[5];
    char sSource[20];
    char sDestination[20];

} ipt_data_t;

int GetDayIndex(uint64_t base_date, uint64_t current_date)
{
#ifdef DEBUG_CODE
    if(current_date > base_date)
    {
        int ii = 0;
    }
#endif

    int iday_index      = 0;
    char szBase[50]     ="";
    char szCurrent[50]  ="";
    char szTemp[5]      ="";

    sprintf(szBase,"%llu",base_date);
    sprintf(szCurrent,"%llu",current_date);

    struct tm base_tm;
    memset(&base_tm,0,sizeof(struct tm));
    base_tm.tm_year = atoi(strncpy(szTemp,&szBase[0],4)) - 1900;
    memset(szTemp,0,5);
    base_tm.tm_mon  = atoi(strncpy(szTemp,&szBase[4],2)) - 1;
    memset(szTemp,0,5);
    base_tm.tm_mday = atoi(strncpy(szTemp,&szBase[6],2));
    memset(szTemp,0,5);

    struct tm current_tm;
    memset(&current_tm,0,sizeof(struct tm));
    current_tm.tm_year = atoi(strncpy(szTemp,&szCurrent[0],4)) - 1900;
    memset(szTemp,0,5);
    current_tm.tm_mon  = atoi(strncpy(szTemp,&szCurrent[4],2)) - 1;
    memset(szTemp,0,5);
    current_tm.tm_mday = atoi(strncpy(szTemp,&szCurrent[6],2));
    memset(szTemp,0,5);

    time_t tNewBaseDate     = mktime(&base_tm);
    time_t tNewCurrentDate  = mktime(&current_tm);

#ifdef DEBUG_CODE
    if(tNewBaseDate < 0 || tNewCurrentDate < 0)
    {
        char szBuf[100]="";
        sprintf(szBuf,"In [%s] Invalid date(s) specified tNewBaseDate [%ld] tNewCurrentDate: [%ld]",__FUNCTION__,tNewBaseDate,tNewCurrentDate);
        perror(szBuf);

    }
#endif

    if(tNewCurrentDate != tNewBaseDate)
    {
        iday_index = ((tNewCurrentDate - tNewBaseDate) / 86400);
    }

#ifdef DEBUG_CODE
    if(iday_index != 0)
    {
        int ii = 0;
    }
#endif

    if(iday_index == -1)
    {
        iday_index = 0;
    }

    return iday_index;
}

void SetDailyBandwidthForHost(int iday_index, double bytes, double bytes_array[], double previous_bytes_array[],int iLogFileIsOnANewDate)
{
    double previous_bytes = previous_bytes_array[iday_index];
    double current_bytes  = bytes_array[iday_index];

    // Get previous date's bandwidth value to carry forward to the date we are working on
    if(iLogFileIsOnANewDate && (iday_index - 1 >= 0))
    {
        previous_bytes = previous_bytes_array[iday_index-1];
    }

    if(bytes > 0 && previous_bytes > 0 && bytes < previous_bytes)
    {
        bytes_array[iday_index] += bytes;
    }
    else if(bytes > 0 && previous_bytes > 0 && previous_bytes < bytes)
    {
        double dAddBytes = (bytes - previous_bytes);
        bytes_array[iday_index] += dAddBytes;
    }
    else if(bytes > 0 && previous_bytes == 0)
    {
        bytes_array[iday_index] += bytes;
    }
#ifdef DEBUG_CODE
    else
    {
        int ii = 0;
        if(bytes != previous_bytes || bytes != current_bytes || current_bytes != previous_bytes)
        {
            int ii2 = 0;
        }
    }
#endif
}

void ProcessIPTableRow( char *pLogFileName,
                        int iDirection,
                        ipt_data_t *pData,
                        data_t ppBandwidthUsage[],
                        int iMaxClientCount,
                        int iLogFileIsOnANewDate)
{
    if(strlen(pData->sDestination) == 0 && strlen(pData->sSource) == 0)
    {
        return;
    }

    char *pHostLookup = NULL;
    if(iDirection == INCOMING_DATA)
    {
        pHostLookup = &pData->sDestination[0];
    }
    else
    {
        pHostLookup = &pData->sSource[0];
    }

#ifdef DEBUG_HOST
    if(strncasecmp(pHostLookup,DEBUG_HOST,strlen(pHostLookup)) != 0)
    {
        return;
    }
#endif

    char *pBytes = pData->sBytes;
    double bytes = atoi(pBytes);

    // Convert bytes to a standard value in MB for consistent calculations
    if(bytes > 0)
    {
        if(strstr(pBytes,"K") != NULL)
        {
            bytes *= 0.001;
        }
        else if(strstr(pBytes,"M") != NULL)
        {

        }
        else if(strstr(pBytes,"G") != NULL)
        {
            bytes *= 1000;
        }
        else
        {
            bytes *= 0.000001;
        }
    }

	int iFirstEmptyIndex    = -1;
	int bFoundClient        = 0;

	int iIdx = 0;
    for(iIdx = 0;
        iIdx < iMaxClientCount;
		iIdx ++)
    {
        data_t *pBandwidthUsage = &ppBandwidthUsage[iIdx];

		if(strncasecmp(pHostLookup,pBandwidthUsage->ip,strlen(pHostLookup)) == 0)
	    {
			bFoundClient = 1;

			char szFileDate[9]  = "";
			int iFileNameLen    = strlen(pLogFileName);
			memcpy(szFileDate,&pLogFileName[iFileNameLen-14],8);

			uint64_t uiFileDate = atoi(szFileDate);
			int iday_index      = GetDayIndex(start_date,uiFileDate);

			if(iday_index >= MAX_RANGE || iday_index < 0)
			{
                char szBuf[100]="";
                sprintf(szBuf,"In [%s] #1 Invalid iday_index value [%d]",__FUNCTION__,iday_index);
                perror(szBuf);
			}

			if(iDirection == INCOMING_DATA)
			{
#ifdef DEBUG_CODE
                char *pPreviousStr    = &pBandwidthUsage->previous_bytes_in_str[0];
#endif
                if(uiFileDate < start_date)
                {
                    pBandwidthUsage->bytes_in[iday_index] = bytes;
                }
                else
                {
                    SetDailyBandwidthForHost(iday_index, bytes, pBandwidthUsage->bytes_in, pBandwidthUsage->previous_bytes_in,iLogFileIsOnANewDate);
                }

				pBandwidthUsage->previous_bytes_in[iday_index] = bytes;

#ifdef DEBUG_CODE
				double dValue = pBandwidthUsage->bytes_in[iday_index];
				if(dValue < 0)
				{
				    int ii = 0;
				}

				strcpy(pBandwidthUsage->previous_bytes_in_str,pBytes);
#endif
			}
			else
			{
#ifdef DEBUG_CODE
			    char *pPreviousStr    = &pBandwidthUsage->previous_bytes_out_str[0];
#endif

                if(uiFileDate < start_date)
                {
                    pBandwidthUsage->bytes_out[iday_index] = bytes;
                }
                else
                {
                    SetDailyBandwidthForHost(iday_index, bytes, pBandwidthUsage->bytes_out, pBandwidthUsage->previous_bytes_out,iLogFileIsOnANewDate);
                }

				pBandwidthUsage->previous_bytes_out[iday_index] = bytes;

#ifdef DEBUG_CODE
				double dValue = pBandwidthUsage->bytes_out[iday_index];
				if(dValue < 0)
				{
				    int ii = 0;
				}

				strcpy(pBandwidthUsage->previous_bytes_out_str,pBytes);
#endif
			}

            if(end_date < uiFileDate)
            {
                end_date = uiFileDate;
            }

			break;
		}
		else if(iFirstEmptyIndex < 0 && pBandwidthUsage->ip[0] == '\0')
		{
			iFirstEmptyIndex = iIdx;
			break;
		}
    }

	if(bFoundClient == 0 && iFirstEmptyIndex >= 0)
	{
		char szFileDate[9]  = "";
		int iFileNameLen    = strlen(pLogFileName);
		memcpy(szFileDate,&pLogFileName[iFileNameLen-14],8);

		uint64_t uiFileDate     = atoi(szFileDate);
		data_t *pBandwidthUsage = &ppBandwidthUsage[iFirstEmptyIndex];

		int iday_index = 0;
		if(start_date > 0)
		{
			iday_index = GetDayIndex(start_date,uiFileDate);
		}

        if(iday_index >= MAX_RANGE || iday_index < 0)
        {
            char szBuf[100]="";
            sprintf(szBuf,"In [%s] #2 Invalid iday_index value [%d]",__FUNCTION__,iday_index);
            perror(szBuf);
        }

		strcpy(pBandwidthUsage->ip,pHostLookup);
		if(iDirection == INCOMING_DATA)
		{
			pBandwidthUsage->bytes_in[iday_index] 	        = bytes;
			pBandwidthUsage->previous_bytes_in[iday_index]  = bytes;

#ifdef DEBUG_CODE
			strcpy(pBandwidthUsage->previous_bytes_in_str,pBytes);
#endif
		}
		else
		{
			pBandwidthUsage->bytes_out[iday_index] 	        = bytes;
			pBandwidthUsage->previous_bytes_out[iday_index] = bytes;

#ifdef DEBUG_CODE
			strcpy(pBandwidthUsage->previous_bytes_out_str,pBytes);
#endif
		}

		if(start_date <= 0)
		{
			start_date = uiFileDate;
		}
		if(end_date < uiFileDate)
		{
			end_date = uiFileDate;
		}
	}
}

void ProcessLogFile(char *szLogPath, char *szLogFile, data_t tBandwidthUsage[],int iDirection, int iLogFileIsOnANewDate)
{
    FILE *f                     = 0;
    char szFullFilePath[1024]   = "";
    sprintf(szFullFilePath,"%s%s",szLogPath,szLogFile);

#ifdef DEBUG_CODE
    printf("Processing file [%s]\n",szFullFilePath);
#endif

    if ((f = fopen(szFullFilePath, "r")) != NULL)
    {
        char szDiscard[1024] = "";
        fgets(szDiscard,1023,f);
        fgets(szDiscard,1023,f);

        while(feof(f) == 0)
        {
            ipt_data_t tData;
            memset(&tData,0,sizeof(tData));

            int iCount = fscanf(f, "%s %s %s %s %s %s %s %s",
                                tData.sPackets,
                                tData.sBytes,
                                tData.sProt,
                                tData.sOp,
                                tData.sIn,
                                tData.sOut,
                                tData.sSource,
                                tData.sDestination);

            if(iCount == 8)
            {
                ProcessIPTableRow(szLogFile, iDirection, &tData,
                                  tBandwidthUsage,MAX_CLIENTCOUNT,iLogFileIsOnANewDate);
            }
        }
        fclose(f);
    }
}

void ProcessTrafficByDirectionFlow( char *szLogPath,
                                    char *szLogFilePrefix,
                                    int iDirection,
                                    char *szLogFileFilterStart,
                                    char *szLogFileFilterEnd,
                                    data_t tBandwidthUsage[],
                                    data_t tBandwidthUsageInitValue[])
{
    DIR *dp = opendir(szLogPath);
    if (dp != NULL)
    {
        char szStartFile[1024]  = "";
        char szEndFile[1024]    = "";

        sprintf(szStartFile,"%s%s",szLogFilePrefix,szLogFileFilterStart);
        sprintf(szEndFile,"%s%s",szLogFilePrefix,szLogFileFilterEnd);

        char szStartFileDate[9]  = "";
        int iStartFileNameLen    = strlen(szLogFileFilterStart);
        memcpy(szStartFileDate,&szLogFileFilterStart[iStartFileNameLen-8],8);
        uint64_t uiStartFileDate = atoi(szStartFileDate);

        char szEndFileDate[9]  = "";
        int iEndFileNameLen    = strlen(szLogFileFilterEnd);
        memcpy(szEndFileDate,&szLogFileFilterEnd[iEndFileNameLen-8],8);
        uint64_t uiEndFileDate = atoi(szEndFileDate);

        int iMaxDateRangeIndex = GetDayIndex(uiStartFileDate, uiEndFileDate);

        int iLastDateRangeIndex = -1;

        int bFoundAndUsedPreviousFile   = 0;
        char szInitFile[1024]           = "";
        struct dirent *ep               = 0;

        uint64_t uiFirstMatchedFileDate = 0;
        uint64_t uiLastMatchedFileDate  = 0;

    	while (ep = readdir(dp))
        {
            // Does the file match the initial filter prefix
            if(strncasecmp(ep->d_name,szLogFilePrefix,strlen(szLogFilePrefix)) == 0)
            {
                char szFileDate[9]  = "";
                int iFileNameLen    = strlen(ep->d_name);
                memcpy(szFileDate,&ep->d_name[iFileNameLen-14],8);
                uint64_t uiFileDate = atoi(szFileDate);

                int iFoundStartFileFilter   = (strncasecmp(ep->d_name,szStartFile,strlen(szStartFile)) == 0);
                int iFoundEndFileFilter     = (strncasecmp(ep->d_name,szEndFile,strlen(szEndFile)) == 0);

                if(uiFirstMatchedFileDate == 0 && iFoundStartFileFilter)
                {
                    uiFirstMatchedFileDate = uiFileDate;
                }
                if(uiLastMatchedFileDate == 0 && iFoundEndFileFilter)
                {
                    uiLastMatchedFileDate = uiFileDate;
                }

                int iTestInDateRange = GetDayIndex(uiStartFileDate, uiFileDate);

                int iIncludeFileInCalcs     = (iFoundStartFileFilter || iFoundEndFileFilter);
                if(!iIncludeFileInCalcs)
                {
                    iIncludeFileInCalcs = (uiFirstMatchedFileDate > 0 && uiFileDate >= uiFirstMatchedFileDate && uiLastMatchedFileDate == 0);

                    if(!iIncludeFileInCalcs)
                    {
                        iIncludeFileInCalcs = (iTestInDateRange >= 0 && iTestInDateRange <= iMaxDateRangeIndex && uiFileDate >= uiStartFileDate);

#ifdef DEBUG_CODE
                        if(iIncludeFileInCalcs)
                        {
                            int ii = 0;
                        }
#endif
                    }
                }

                if(iIncludeFileInCalcs)
                {
                    ProcessLogFile(szLogPath, ep->d_name, tBandwidthUsage,iDirection,(iLastDateRangeIndex != iTestInDateRange));
                    iLastDateRangeIndex = iTestInDateRange;

                    if(bFoundAndUsedPreviousFile == 0)
                    {
                        bFoundAndUsedPreviousFile = 1;

                        if(strlen(szInitFile) > 0)
                        {
                            ProcessLogFile(szLogPath, szInitFile, tBandwidthUsageInitValue,iDirection,1);
                        }
                    }
                }
                else if(bFoundAndUsedPreviousFile == 0)
                {
                    strcpy(szInitFile,ep->d_name);
                }
            }
        }

        (void) closedir(dp);
    }
    else
    {
        perror("Couldn't open the directory");
    }
}

void ProcessSpecialDateParameter(char *pLogFileFilter)
{
    time_t now = time ( NULL );
    struct tm *tm_now = localtime ( &now );

    if(strlen(pLogFileFilter) == 0)
    {
        strftime (pLogFileFilter, 1023, "%Y%m%d", tm_now);
    }
    else if(strstr(pLogFileFilter,"today") != NULL)
    {
        char *pNumericValue = strstr(pLogFileFilter,"-");
        if(pNumericValue != NULL)
        {
            pNumericValue++;
            int iDecrementValue = atoi(pNumericValue);
            tm_now->tm_mday     -= iDecrementValue;
            time_t tNewDate     = mktime(tm_now);
            if(tNewDate < 0)
            {
                char szBuf[100]="";
                sprintf(szBuf,"Invalid logfilter specified [%s] returned: [%ld]",pLogFileFilter,tNewDate);
                perror(szBuf);
            }

            tm_now = localtime ( &tNewDate );
        }
        strftime (pLogFileFilter, 1023, "%Y%m%d", tm_now);
    }
}

int main(int argc, char *argv[])
{
    if (argc < 2)
    {
        printf("ipt-parse\nCopyright (C) 2009 Mark Vejvoda\n\n");
		printf("Usage: <generate-type> <start-date-filter> <end-date-filter> <logfiles-path> <incoming-data-logfileprefix> <outgoing-data-logfileprefix>\n\n");
		printf(": <generate-type>\t= the type of output format to generate (0 = console, 1 = html).\n");
		printf(": <start-date-filter>\t= the optional start date range of logfiles to parse (example: 20090425 and default is current date. You may also use the today keyword like this: today).\n");
		printf(": <end-date-filter>\t= the optional end date range of logfiles to parse (example: 20090425 and default is current date. You may also use the today keyword like this: today-30).\n");
		printf(": <logfiles-path>\t= the optional path to search for the logfiles (default value is ./).\n");
		printf(": <incoming-data-logfileprefix>\t= the optional log filename prefix used to describe incoming data logs (example: traffic_in_).\n");
		printf(": <outgoing-data-logfileprefix>\t= the optional log filename prefix used to describe outgoing data logs (example: traffic_out_).\n");
		return -1;
    }

    time_t tStartProcessing = time(NULL);

    int iGenerateType = atoi(argv[1]);
    if(iGenerateType == 0)
    {
        printf("ipt-parse\nCopyright (C) 2009 Mark Vejvoda\n\n");
    }

    data_t tBandwidthUsage[MAX_CLIENTCOUNT];
    data_t tBandwidthUsageInitValue[MAX_CLIENTCOUNT];
    memset(&tBandwidthUsage[0],0,sizeof(data_t) * MAX_CLIENTCOUNT);
    memset(&tBandwidthUsageInitValue[0],0,sizeof(data_t) * MAX_CLIENTCOUNT);

    char szLogPath[1024]="./";
    if (argc >= 5)
    {
        sprintf(szLogPath,"%s",argv[4]);
    }

    char szIncomingDataLogfilePrefix[1024]="";
    if (argc >= 6)
    {
        sprintf(szIncomingDataLogfilePrefix,"%s",argv[5]);
    }

    char szOutgoingDataLogfilePrefix[1024]="";
    if (argc >= 7)
    {
        sprintf(szOutgoingDataLogfilePrefix,"%s",argv[6]);
    }

    char szLogFileStartFilter[1024]="";
    if (argc >= 3)
    {
        sprintf(szLogFileStartFilter,"%s",argv[2]);
    }
    ProcessSpecialDateParameter(szLogFileStartFilter);

    char szLogFileEndFilter[1024]="";
    if (argc >= 4)
    {
        sprintf(szLogFileEndFilter,"%s",argv[3]);
    }
    ProcessSpecialDateParameter(szLogFileEndFilter);

    if(iGenerateType == 0)
    {
        printf("Using filter date-range [%s] - [%s]\n",szLogFileStartFilter,szLogFileEndFilter);
    }

    ProcessTrafficByDirectionFlow( szLogPath, szIncomingDataLogfilePrefix, INCOMING_DATA,
                                   szLogFileStartFilter,szLogFileEndFilter, tBandwidthUsage,tBandwidthUsageInitValue);

    ProcessTrafficByDirectionFlow( szLogPath, szOutgoingDataLogfilePrefix, OUTGOING_DATA,
                                   szLogFileStartFilter,szLogFileEndFilter, tBandwidthUsage,tBandwidthUsageInitValue);


    if(iGenerateType == 0)
    {
        printf("=====================================================================================================\n");
        printf("Bandwidth usage from: %llu to %llu\n",start_date,end_date);
    }
    else if(iGenerateType == 1)
    {
        printf("\
        <html>\
        <link rel='stylesheet' type='text/css' href='tomato.css'>\
        <link rel='stylesheet' type='text/css' href='red.css'>\
        <body>\
        <b>Bandwidth usage from: %llu to %llu</b>\
        <table class=\"bwmg\" cellspacing=\"1\">\
        <tbody>\
        <tr class=\"header\">\
        <td class=\"rtitle\">Host</td>\
        <td class=\"dl\">Download</td>\
        <td class=\"ul\">Upload</td>\
        <td class=\"total\">Total</td>\
        </tr>",start_date,end_date);
    }

    double master_bytes_in_total  = 0;
    double master_bytes_out_total = 0;

	int iIdx = 0;
    for(iIdx = 0;
	    iIdx < MAX_CLIENTCOUNT;
		iIdx ++)
    {
        data_t *pBandwidthUsage     = &tBandwidthUsage[iIdx];
        //data_t *pBandwidthUsageInit = &tBandwidthUsageInitValue[iIdx];

        if(pBandwidthUsage->ip[0] == '\0') break;

        double bytes_in_total  = 0;
        double bytes_out_total = 0;

        int iIdx2 = 0;
        for(iIdx2 = 0;
            iIdx2 < MAX_RANGE;
            iIdx2 ++)
        {
            bytes_in_total  += pBandwidthUsage->bytes_in[iIdx2];
            bytes_out_total += pBandwidthUsage->bytes_out[iIdx2];
        }

        int iIdx3 = 0;
        for(iIdx3 = 0;
            iIdx3 < MAX_CLIENTCOUNT;
            iIdx3 ++)
        {
            data_t *pBandwidthUsageInit = &tBandwidthUsageInitValue[iIdx3];

            if(strcasecmp(pBandwidthUsageInit->ip,pBandwidthUsage->ip) == 0)
            {
                // Subtract off previous days start values
                double dInitialBytesIn  = pBandwidthUsageInit->bytes_in[0];
                double dInitialBytesOut = pBandwidthUsageInit->bytes_out[0];

                //if(dInitialBytesIn < bytes_in_total)
                {
                    bytes_in_total  -= dInitialBytesIn;
                }
                //if(dInitialBytesOut < bytes_out_total)
                {
                    bytes_out_total -= dInitialBytesOut;
                }

                break;
            }
        }

        master_bytes_in_total += bytes_in_total;
        master_bytes_out_total += bytes_out_total;

        if(iGenerateType == 0)
        {
            printf("Totals for host [%s]\tIncoming [%.2f MB]\tOutgoing [%.2f MB]\tTotal [%.2f MB]\n",pBandwidthUsage->ip,bytes_in_total,bytes_out_total,(bytes_in_total + bytes_out_total));
        }
        else if(iGenerateType == 1)
        {
            printf("\
            <tr class=\"%s\">\
            <td class=\"rtitle\">%s</td>\
            <td class=\"dl\">%.2f MB</td>\
            <td class=\"ul\">%.2f MB</td>\
            <td class=\"total\">%.2f MB</td>\
            </tr>",
            (iIdx % 2 == 0 ? "even" : "odd"),
            pBandwidthUsage->ip,
            bytes_in_total,
            bytes_out_total,
            (bytes_in_total + bytes_out_total));
        }

    }

    double dElapsedSeconds = difftime(time(NULL),tStartProcessing);

    if(iGenerateType == 0)
    {
        printf("=====================================================================================================\n");
        printf("Totals for all hosts in filter\tIncoming [%.2f MB]\tOutgoing [%.2f MB]\tTotal [%.2f MB]\n",master_bytes_in_total,master_bytes_out_total,(master_bytes_in_total + master_bytes_out_total));
        printf("\nElapsed time to generate stats: %.2f\n",dElapsedSeconds);
    }
    else if(iGenerateType == 1)
    {
        printf("\
        <tr class=\"%s\">\
        <td class=\"rtitle\">%s</td>\
        <td class=\"dl\">%.2f MB</td>\
        <td class=\"ul\">%.2f MB</td>\
        <td class=\"total\">%.2f MB</td>\
        </tr>",
        "header", //(iIdx % 2 == 0 ? "even" : "odd"),
        "Total",
        master_bytes_in_total,
        master_bytes_out_total,
        (master_bytes_in_total + master_bytes_out_total));

        printf("\
        </tbody>\
        </table>\
        </html>\
        </body>\
        ");
    }

    return 0;
}
